#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Created on Fri Aug  9 10:40:27 2019
    Use an ICOS digital object id (persistent identification url)
    to load a binary representation of the data object.
"""

__author__      = ["Claudio D'Onofrio"]
__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.3"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "2020-10-16"

import os
import requests
import struct
import pandas as pd
import re

from icoscp.sparql.runsparql import RunSparql
import icoscp.sparql.sparqls as sparqls
from icoscp.cpb import dtype_dict

class Dobj():
    """ Use an ICOS digital object id to query the sparql endpoint
        for infos, and create the "payload" to retrieve the binary data
        the method .getColumns() will return the actual data
    """
    
    def __init__(self, digitalObject = ''):          
        
        self._colSelected = None    # 'none' -> ALL columns are returned
        self._endian = 'big'        # default "big endian conversion 
        self._dtconvert = True      # convert Unixtimstamp to datetime object
        self._info1 = None          # sparql query object           
        self._info2 = None          # sparql query objcect format
        self._info3 = None          # station information for dobj
        self._colSchema = None      # format of columns (read struct bin data)        
        self._dtypeStruct = None    # format of columns (read struct bin data)        
        self._json = None           # holds the "payload" for requests        
        self._server = 'https://data.icos-cp.eu/portal/tabular'
        self._localpath = '/data/dataAppStorage/'        # read data from local file
        self._islocal = None        # status if file is read from local store
                                    # if localpath + dobj is valid                                     
        self._dobjSet = False       # -> see dobj setter
        self._dobjValid = False     # -> see __getPayload()
        
        self.citation = None        # hold the citation string for the object
        
        self._data = pd.DataFrame() # This holds the data pandas dataframe
                                    # persistent in the object. 
        self._datapersistent = True # If True (default), data is kept persistent
                                    # in self._data. If False, force to reload
                                    # from data portal.
        
        # this needs to be the last call within init. If dobj is provided
        # __payLoad() is exectued automatically to create json object 
        # for all columns#
        self.dobj = digitalObject   # this sets self._dobj, which is the PID        
        
        
    #-----------
    @property
    def dobj(self):
        return self._dobj
    
    @dobj.setter
    def dobj(self, digitalObject=''):
        if not digitalObject:
            self._dobjSet = False
            return
        else:
            self._dobj = digitalObject
            self._dobjSet = True
            self.__getPayload()
    #-----------
    @property
    def id(self):
        return self._dobj
    
    @id.setter
    def id(self, id=''):
        self.dobj = id
    #-----------
    @property
    def valid(self):
        return self._dobjValid
    #-----------
    @property
    def dateTimeConvert(self):
        return self._dtconvert
    #-----------
    @dateTimeConvert.setter
    def dateTimeConvert(self, convert=True):
        self._dtconvert = convert
    #-----------
    @property
    def colNames(self):
        if self._dobjValid:
            return self._info2['colName']
        return None  
    #-----------
    @property
    def station(self):
        return self._info3.stationName[0]
    #-----------
    @property
    def lat(self):
        return float(self._info3.latitude[0])
    #-----------
    @property
    def lon(self):
        return float(self._info3.longitude[0])
    #-----------
    @property
    def elevation(self):
        return float(self._info3.elevation[0])
    #-----------
    @property
    def data(self):
        return self.getColumns()
    #-----------
    @property
    def info(self):
        if self._dobjValid:
            return [self._info1, self._info2, self._info3]
        return None 
    #-----------

# -------------------------------------------------    
    def get(self):
        """ return data as pandas data frame"""
        return self.getColumns()
    
    def getColumns(self, columns=None):
        #check if the data is alreday present and persistent is true.
        # return the existing dataframe in self._data        
        if not self._data.empty and self._datapersistent:
            return self._data

        # if datapersistence is fals or the data is missing download..        
        # if columns = None, return ALL columns, otherwise, 
        # try to extract only a subset of columns
        if (not columns) & self._dobjValid:
            # return all columns
            return self.__getColumns()
        
        if columns:
            if self.__setColumns(columns):
                # set columns and recompute payload
                self.__getPayload()
        
        if self._dobjValid:
            return self.__getColumns()
        
        return False
    
# -------------------------------------------------    
    def __getPayload(self):
        """ this function sends predefined sparql queries to the cp endpoint
            a single digital object is interrogated to check
            if a binary data representation is available
            if successful _dobjValid is "True"
        """
        if not self._dobjSet:
            return
                
        sparql = RunSparql()
        """
        get information about the digital objcect and if 
        there is a dataset with format specification
        It is possible, that "optional" columns are present.
        -   check if columnNames is empty or not. If not adjust 
            info2 accordingly
        """
        
        sparql.query = sparqls.cpbGetInfo(self.dobj)
        sparql.format = 'pandas'
        sparql.run()
        self._info1 = sparql.data()
        if not len(self._info1):
            self._dobjValid=False              
            return

        # -------------------------------------------------
        # get the information about the underlying format
        # from the previous query -> sparql.data....
        # info2 will contain a list of all columns
        sparql.query = sparqls.cpbGetSchemaDetail(self._info1['objSpec'][0])
        sparql.run()  
        self._info2 = sparql.data()
        
        # if info[1][columnNames] contains a list (optional columns...)
        # we need to deal with possible regex entries
        
        if not self._info1.columnNames.empty:
            
            # check if a there is regex column, if yes, deal with it             
            if self._info2.isRegex.count():                
           
                colList = self._info1['columnNames'][0]
                # remove [] from beginning and end
                colList = colList[1:len(colList)-1]
                # remove all double quotes and make a list
                colList = colList.replace('"', '').split(',')
                #remove leading and trailing whitespaces from entries
                colList = [c.strip() for c in colList]
                
                #replace _info1.columnNames with a clean list
                self._info1.loc[0].columnNames = colList
                
                # cycle through all column definition which are regex
                regexrow = self._info2[self._info2.isRegex == 'true']
                for i, r in enumerate(regexrow.colName):
                    # for each column, test if regex matches and
                    # create an entry in self._info2
                    for c in colList:
                        if re.match(r, c):
                            row = regexrow.iloc[[i]]
                            row.loc[row.index[0]].colName = c
                            row.loc[row.index[0]].isRegex = None                            
                            self._info2 = pd.concat([self._info2, row],ignore_index=True)
                
                # now we have expanded all regex column...remove the original regex rows
                self._info2 = self._info2[self._info2.isRegex != 'true']

        # make sure that _info2 is sorted....the binary file representation
        # has been crated that way, hence we need to have the same order
        self._info2.sort_values(by='colName', inplace=True)
            
        # if no specific columns are selected, default to all
        if not self._colSelected:
            self._colSelected = list(range(0,len(self._info2)))

        # -------------------------------------------------
        # create the dataType format, neccesary to interpret the binary 
        # data with python struct library or numpy        
        self._colSchema = []        
        for f in self._info2['valFormat']:
            try:
                self._colSchema.append(dtype_dict.mapDataTypesCP(f))  
            except:
                raise Exception('dtype_dict')
                    
        dtypeStruct = []
        for d in self._colSelected:    
            dtypeStruct.append(dtype_dict.structTypes(self._colSchema[d],int(self._info1['nRows'][0])))                    
            
        self._dtypeStruct = dtype_dict.structEndian(self._endian) + ''.join(dtypeStruct)
        
        # assemble the json object needed to return the data
        self._json = {'tableId':self._info1['dobj'].iloc[0].split('/')[-1],
                          'schema':{'columns':self._colSchema,
                                    'size':int(self._info1['nRows'].iloc[0])},
                                    'columnNumbers':self._colSelected,
                                    'subFolder':self._info2['objFormat'].iloc[0].split('/')[-1]
                     }
        
        # get the citation for this object
        sparql.query = sparqls.get_icos_citation(self.dobj)                   
        citation = sparql.run()['cit'][0] 
        if not citation:
            self.citation = 'no citation available ...'
        else:
            self.citation = citation  
        
        self._dobjValid = True
        return
        
# -------------------------------------------------    
    def __getColumns(self):
        """
            check if a local path is set and valid
            otherwise try to download from the cp server
        """ 
        # get the station information for the dobj, 
        sparql = RunSparql()
        sparql.format = 'pandas'
        sparql.query = sparqls.dobjStation(self.dobj)
        sparql.run()  
        self._info3 = sparql.data()        
        
        
        #assemble local file path        
        folder = self._info2['objFormat'].iloc[0].split('/')[-1]        
        fileName = ''.join([self.dobj.split('/')[-1],'.cpb'])
        localfile = os.path.abspath(''.join([self._localpath,folder,'/',fileName]))
        
        if os.path.isfile(localfile): 
            self._islocal = True
            with open(localfile, 'rb') as binData:
                content = binData.read()
            # we need to select ALL columns
            self._colSelected = None             
            self.__getPayload()
            return self.__unpackRawData(content)  
            
        else:
            self._islocal = False
            r = requests.post(self._server, json=self._json, stream=True) 
            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise Exception(e)

            #track data usage
            self.__portalUse()
            
            return self.__unpackRawData(r.content)                    

    def __unpackRawData(self, rawData):
        # unpack the binary data
        data = struct.unpack_from(self._dtypeStruct, rawData)
        
        # get them into a pandas data frame, column by column
        df = pd.DataFrame()
        rows = int(self._info1['nRows'])  
        
        try:
            for idx,col in enumerate(self._colSelected):                       
                lst = list(data[idx*rows:(idx+1)*rows])
                if(self._colSchema[col] == 'CHAR'):
                    #convert UTF-16 , this is often used in "Flag" columns
                    lst = [chr(i) for i in lst]            
                df[self._info2.iloc[col]['colName']] = lst
        except Exception as e:
            raise Exception('_unpackRawData')
            print(e)
                
        """
            The ICOS Carbon Portal provides a TIMESTAMP which is
            a unix timestamp in milliseconds [UTC]
            Convert the "number" to a pandas date/time object       
        """
        if 'TIMESTAMP' in df.columns and self._dtconvert:
            df['TIMESTAMP'] = pd.to_datetime(df.loc[:,'TIMESTAMP'],unit='ms')
        
        # there are files with date and time where
        # date is a unixtimestamp
        # time seconds per day
        if 'date' in df.columns and self._dtconvert:
            df['date'] = pd.to_datetime(df.loc[:,'date'],unit='D')
        
        if 'time' in df.columns and self._dtconvert:
            # convert unix timestamp
            df['time'] = pd.to_datetime(df.loc[:,'time'],unit='s')
            # remove the data, so that only time remains
            df['time'] = pd.to_datetime(df.loc[:,'time'],format='%H:%M').dt.time
            
        # reorder the data frame in case of regex/optional columns
        # to the original order provided by cpmeta:hasColumns
        if self._info1.loc[0].columnNames:
            df = df[self._info1.loc[0].columnNames]
        
        # store data in object
        if self._datapersistent:
            self._data = df
            
        return df

    # -------------------------------------------------                       
    def __portalUse(self):
        
        """ private function to track data usage """        
        counter = {'BinaryFileDownload':{
        'params':{
            'objId':self._dobj,
            'columns': self.colNames.tolist(),            
            'library':__name__, 
            'version':__version__}
            }
        }
        server = 'https://cpauth.icos-cp.eu/logs/portaluse'        
        requests.post(server, json=counter)
        
    # -------------------------------------------------                       
    def __setColumns(self, columns=None):
        """ this function sets the columnNumbers to extract data 
            :Param colums (None | List)
            
            1. If columns=None returns all ColumnNumbers
            2. List -> either a list of numbers(indices for columns), columNames
                as strings or a mixture of both.
                Only valid uniqe entries will be returned
                
            Return: Bool (True|False)
                    False if:   _dobjValid is False
                                columns is not a list
                                column could not be found in dobj
                    True if self._colSeleceted contains valid entries and hence
                    the "__getPayload" is valid, such that __getData can be exectuded.            
        """
        if not self._dobjValid:
            return False
        
        if columns is None:
            # with colSelected None, payload returns ALL available columns
            self._colSelected = None            
            return True

        if not isinstance(columns, list):
            try:
                columns = list(columns)
            except: 
                return False
        
        colSelected = []
        
        # we deal everything in uppercase to make it case INsensitive
        colNames = [c.upper() for c in self.colNames.tolist()]
        colIndexRange = range(len(colNames))
        for c in columns:
            #check if list entry is a "name"
            if (isinstance(c,str) and c.upper() in colNames):         
                idx = colNames.index(c.upper())                
            # check if an "index" is provided
            elif (isinstance(c, int) and c in colIndexRange):
                idx = c            
            if (idx not in colSelected):
                colSelected.append(idx)
        
        if not colSelected: # the list is empty..            
            return False
        else:
            self._colSelected = colSelected
            
        return True
     
    # ------------------------------------------------------------    
    def size(self):
        """
        return the real size of object
        https://goshippo.com/blog/measure-real-size-any-python-object/
        """
        import icoscp.cpb.get_size as s
        return s.get(self)
    
if __name__ == "__main__":
    """
    execute only if run as a script
    """
    
    msg="""
    You should use this Class within a script.
    Example to create a digital object representation
    
    from icoscp.cpb.dobj import Dobj
    
    do = Dobj('https://meta.icos-cp.eu/objects/M6XCOcBsPDTnlUv_6gGNZ2EX')
    data = do.get() 
    data.head(3)           
    """
       
    print(msg)
    


"""
    The cpb module gives you access to a binary representatin of 
    digital data objects hosted at the ICOS CarbonPortal.
    Any data column (x,y axis) you can "preview" on the data portal
    ( https://data.icos-cp.eu/portal/ ) is accessible from this module.
    You need to know the digital object ID (URL, persistent identification)
    or have a sqparql query providing this information.    
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The collection module is used to explore ICOS data collected for a specific
    project or data product assembly.

    Example usage:

    from icoscp.collection import collection
    collection.getIdList() # returns a pandas data frame with all collections
    
    myCollection = station.get('collectionId') # create a single collection object
    myCollection.info # print information (attributes) about the collection
    myCollection.data # return a list dobj, associated with the collection
    
"""

__author__      = ["Claudio D'Onofrio"]
__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.0"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "20209-09-23"

from icoscp.sparql.runsparql import RunSparql
from icoscp.sparql import sparqls
from icoscp.cpb.dobj import Dobj
from tqdm import tqdm
import pandas as pd
import requests


# ----------------------------------------------
class Collection():
    """ Create an ICOS collection object. This data structure should not be
        instantiated on its own. Please use the collection.get(id) function.
        
        Thec collection object contains the following attributes:
            id: (str) URL Landing Page
            title: (str) Title of the collection
            description: (str) A short description about the collection            
            data: list[dobj] A list of digital data objects, see module cpb.Dobj
            datalink: list[str] A list of PID/URI for all associated datasets
            citation: (str) The citatin string, IF a DOI is associated.
    """
    
    def __init__(self, coll):
        
        self._id = None             # PID, landing page link        
        self._doi = None            # DOI
        self._citation = None       # Citation string IF DOI is available
        self._title = None          # title for the collection
        self._description = None    # description
        self._info = None           # keep the original pandas data frame 
        self._data = None           # a list of dataobjects
        self._datalink = None       # a list of PID's linking to dobj
        
        self.__set__(coll)          # initialize the object with the
                                    # provided pandas dataframe
                                    
    #-------------------------------------
    @property
    def id(self):
        return self._id
    
    @id.setter
    def id(self, id):
        self._id = id
    #-------------------------------------
    @property
    def doi(self):
        return self._doi
    
    @doi.setter
    def doi(self, doi):
        self._doi = doi
    #-------------------------------------
    @property
    def title(self):
        return self._title
    
    @title.setter
    def title(self, title):
        self._title = title
    #-------------------------------------
    @property
    def description(self):
        return self._description
    
    @description.setter
    def description(self, description):
        self._description = description
     #-------------------------------------
    @property
    def data(self):
        return self._data
     #-------------------------------------
    @property
    def datalink(self):
        return self._datalink
    #-------------------------------------
    @property
    def citation(self):
        return self._citation
    #-------------------------------------

    def __str__(self):          
        return self._title
    
    def __set__(self, coll):
        
        self._id = coll.collection.values[0]
        self._doi = coll.doi.values[0]
        self._title = coll.title.values[0]
        self._description = coll.description.values[0]
        
        # get data objects for the collection
         
        query = sparqls.collection_items(self._id)
        dolist = RunSparql(query,'pandas').run()
        
        # keep the list of dobj associated with collection
        self._datalink = dolist.dobj.to_list()
        
        # create dobj objects 
        self._data = []
        for dobj in tqdm(self._datalink):
            self._data.append(Dobj(dobj))     
            
        # citation..        
        self._citation = self.getCitation()
        
        # create the info as dictionary
        self._info = {
            "id": self._id,
            "doi": self._doi,
            "citation" : self._citation,
            "title" : self._title,
            "description" : self._description
            }

    def info(self, fmt='dict'):
        """
        Return information about the collection
        (id, doi, citation, title, description).

        Parameters
        ----------
        fmt : STR, optional, define output format:
            ["dict" | 'pandas' | 'html']. The default is 'dict'.

        Returns
        -------
        FMT
        """
    
        if fmt=='dict':
            return self._info
        
        # convert dict to pandas dataframe....use pandas built-in 
        # converters afterwards
        df = pd.DataFrame.from_dict(self._info, orient='index')
        if fmt=='pandas':
            return df
        if fmt=='html':
            return df.to_html()
        else:
            # default if format is not recogized
            return self._info
    

    def getCitation(self, format='apa', lang='en-GB'):
        '''
        Get a citation string from https://citation.crosscite.org/
        You can provide style & language parameters, default is a simple text
        representation, which is stored in the attribute collection.citation
        
        Example to get a Bibtex styled citation :
            getCitation('bibtex','en-US')
        
        Parameters
        ----------
        format : STR, optional. The default is 'apa'.
        lang : STR, optional. The default is 'en-GB'.
    
        Returns
        -------
        STR, Citation for DOI
    
        '''
        # check if a doi is available at all
        if not self._doi:
            self._citation = 'no citation available'
            return self._citation
        
        # get the citation 
        header = {'accept': 'text/plain'}        
        url = 'https://citation.crosscite.org/format?'
        
        doi = ''.join(['doi=',self._doi])
        style = ''.join(['&style=', format])
        
        lang = ''.join(['&lang=', 'en-GB'])
        query = ''.join([url,doi, style, lang])
        
        r = requests.get(query)
        return(r.content.decode('UTF-8'))

               
# --EOF Collection Class-----------------------------------------            
# ------------------------------------------------------------
def get(collectionId):
    """
    Parameters
    ----------
    collectionId : str StationId (the PID for the collection)
                    extract all collection Id's with getIdList()
                
    Returns
    -------
    collectionn : object Returns a collectionn object, containing the meta
                    data for the collection and a list of dobj.

    """
   
    query = sparqls.collections(collectionId)    
    coll = RunSparql(query,'pandas').run()
 
    # make sure we have one and only one collection
    
    if len(coll)==1:
        return Collection(coll)
    else:
        msg = 'No collection found. Check Id.'
        return msg, False

    
def getIdList():
    """
    Returns a list with all known collections.
    A collection is an assembly of data objects. Have a look at
    https://www.icos-cp.eu/data-products

  
    Returns
    -------
    Pandas data frame with collection Id's, DOI's and Title,...
    """    
    
    query = sparqls.collections()
    coll = RunSparql(query,'pandas').run()
    
    # register tqdm with pandas then change .apply to .progress_apply
    # would be nice, but creates warning with future ...
    # tqdm.pandas()    
    
    # add count of data sets
    coll['dobj'] = coll.apply(lambda x : __itemCount(x['collection']), axis=1)
    coll['count'] = coll.apply(lambda x : len(x.dobj), axis=1)
    
    return coll

def __itemCount(id):
    query = sparqls.collection_items(id)
    item = RunSparql(query,'pandas').run()
    return item.dobj.to_list()
    
# ------------------------------------------------------------    
if __name__ == "__main__":
    """
    execute only if run as a script
    """
    
    msg="""
    Example to get a list of all available collections
    from icoscp.collection import collection
    collection.getIdList()
        
    Example to get a collection and print the citation
    """ 
    c = get('https://meta.icos-cp.eu/collections/n7cIMHIyqHJKBeF_3jjgptHP')
    print(c.info('pandas'))
    #print(msg)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""    
    The Station module is used to explore ICOS stations and the corresponding
    data products. Since you need to know the "station id" to create a station
    object, convenience functions are provided:
    
    Example usage:
        
    from icoscp.station import station    
    station.getIdList() # returns a pandas data frame with all station Id's'
    myStation = station.get('StationId') # create a single statino object
    myList = station.getList('AS') #  returns a list of Atmospheric stations
"""

__author__      = ["Claudio D'Onofrio"]
__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.1"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "2019-08-09"

import json
import pandas as pd
from icoscp.sparql.runsparql import RunSparql
from icoscp.sparql import sparqls
from tqdm import tqdm

# ----------------------------------------------
class Station():
    """ Create an ICOS Station object. This class intends to create 
        an instance of station, providing meta data including 
        stationId, Name, PI, Lat, Lon,  etc.
        Examples:
        import station
        myList = station.getList(['AS']) 
        myStation = station.get('HTM')
        
        station.info()
        
        # extract single attribute
        station.lat -> returns latitude as float
    """
    
    def __init__(self):
        """
        Initialize your Station either with NO arguments, or
        provide a list of arguments in the exact order how the
        attributes are listed 
        [ stationID | stationName | theme | class | type | lat |
         lon | eas | eag | firstname | lastname | email | country ]
      
        """

        # Be aware, that attrList needs to be in the same ORDER as attributes                
        # info
        self._stationId = None      # shortName like HTM or SE-NOR   
        self._valid = False         # if stationId is set and valid, return True            
        self._name = None           # longName
        self._theme = None          # AS | ES | OS
        self._icosclass = None      # 1 | 2 
        self._siteType = None       # description of site
        
        #locations
        self._lat = None            # latitude
        self._lon = None            # longitude
        self._eas = None            # elevation above sea level
        self._eag = None            # elevation above ground, height of tower
        
        #pi information
        self._firstName = None      # Station PI first name
        self._lastName = None       # Station PI last name
        self._email = None          # Station PI email
        
        # other information
        self._country = None
        self._project = None        # list, project affiliation, 
        self._uri = None            # list, links to ressources, landing pages
        
            
    #super().__init__() # for subclasses
    #-------------------------------------
    @property
    def stationId(self):
        return self._stationId
    
    @stationId.setter
    def stationId(self, stationId):
        self._stationId = stationId
    #-------------------------------------
    @property
    def valid(self):
        return self._valid
    
    @valid.setter
    def valid(self, valid):
        self._valid = valid
    #-------------------------------------
    @property
    def theme(self):
        return self._theme
    
    @theme.setter
    def theme(self, theme):
        self._theme = theme
     #-------------------------------------
    @property
    def icosclass(self):
        return self._icosclass
    
    @icosclass.setter
    def icosclass(self, icosclass):
        self._icosclass = icosclass
     #-------------------------------------    
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, name):
        self._name = name    
    #-------------------------------------
    @property
    def lat(self):
        return self._lat
    
    @lat.setter
    def lat(self, lat):
        self._lat = lat    
    #-------------------------------------
    @property
    def lon(self):
        return self._lon
    
    @lon.setter
    def lon(self, lon):
        self._lon = lon    
    #-------------------------------------
    @property
    def eas(self):
        return self._eas
    
    @eas.setter
    def eas(self, eas):
        self._eas = eas    
    #-------------------------------------
    @property
    def eag(self):
        return self._eag
    
    @eag.setter
    def eag(self, eag):
        self._eag = eag    
    #-------------------------------------
    @property
    def firstName(self):
        return self._firstName
    
    @firstName.setter
    def firstName(self, firstName):
        self._firstName = firstName    
    #-------------------------------------
    @property
    def lastName(self):
        return self._lastName
    
    @lastName.setter
    def lastName(self, lastName):
        self._lastName = lastName
    #-------------------------------------
    @property
    def email(self):
        return self._email
    
    @email.setter
    def email(self, email):
        self._email = email
    #-------------------------------------     
    @property
    def country(self):
        return self._country
    
    @country.setter
    def country(self, country):
        self._country = country
    #-------------------------------------
    @property
    def project(self):
        return self._project
    
    @project.setter
    def project(self, project):
        self._project = project
    #-------------------------------------
    @property
    def uri(self):
        return self._uri
    
    @uri.setter
    def uri(self, uri):
        self._uri = uri
    #-------------------------------------

    def __str__(self):          
        return self.info('json')
       
        
    def data(self, level=None):
        """
        return a list of digital objects for the station
        parameter: level [str | int] default None
        provide filter to dataobjects
            1 = raw data
            2 = QAQC data
            3 = elaborated products
        """
        if not isinstance(self._data, pd.DataFrame):
            # _data is not a dataframe but contains a string...
            return self._data

        if level:
            # if level is provided, filter pandas data frame
            return self._data[self._data['datalevel']==str(level)]
        else:
            # return complete pandas data fram
            return self._data
    
    def products(self, fmt='pandas'):
        """
        Parameters
        ----------
        fmt : str,  optional ['pandas', 'dict']
                    Return a pandas dataframe of unique data products for
                    the station. The default is 'pandas'.

        Returns
        -------
        [pandas data frame | dict]
            uniqe list of data products.

        """
        
        if not isinstance(self._products, pd.DataFrame):
            return self._products
        
        if fmt == 'dict':
            return self._products.to_dict()
        else:
            return self._products
    
    
    def setStation(self, attrib=None):
        """
        You can set attributes for the station by providing a dictionary        
        Parameters.      
        project & uri must be a list
        lat, lon, eas, eag, must be convertible to float
        everything else is a string.
        
        ----------
        attributeList : dictionary containing attributes
        
        - stationId (shortName)
        - name (longName)
        - theme (AS, ES, OS)
        - Class (ICOS Station class, see ICOS Handbook)
        - siteType (description of site, if available)
        - latitude (convertible to float)
        - longitude (convertible to float)
        - eas (elevation above sea level) (convertible to float)
        - eag (elevation above ground, height of the tower) (convertible to float)   
        - firstName (PI first name)
        - lastName (PI last name)
        - email (PI email)    
        - country                    
        
        Returns
        -------
        None
        """
        
        if not isinstance(attrib, dict):            
            return
        
        # minimal sanity check
        checkFloat = ['lat','lon','eag','eas']
        checkList = ['project', 'uri']
                                
        #create 'keys' without the underscore
        keys = [k.strip('_') for k in list(self.__dict__)]        
        for a in attrib:
            a = a.strip('_')
            if a in keys and a in checkFloat:
                try:
                    self.__setattr__('_'+a, float(attrib[a]))
                except:
                    continue
            elif a in keys and a in checkList:
                try:
                    self.__setattr__('_'+a, list(attrib[a]))
                except:
                    continue
            else:
                self.__setattr__('_' + a, attrib[a])
                


    def info(self, fmt = 'dict'):
        """
        Parameters
        ----------
        fmt : str ['dict' | 'json' | 'list' | 'pandas' | 'html']
            You can choose a return format by providing fmt.
            The default is 'dict'.
            

        Returns
        -------
        All attributes for the Station

        """
        #create new 'keys' without the underscore
        newKeys = [k.strip('_') for k in list(self.__dict__)]
        values =  self.__dict__.values()
        dictionary = dict(zip(newKeys,values))
        
        # remove data and products from the dict to get a shorter
        # summary of information about station
        if 'data' in dictionary:
            del dictionary['data']
            
        if 'products' in dictionary:
            del dictionary['products']
            
        if 'valid' in dictionary:
            del dictionary['valid']
        
        if fmt == 'dict':
            return dictionary
        
        if fmt == 'json':
            return json.dumps(dictionary, indent=4)
                        
        if fmt == 'list':
            return (list(dictionary.keys()), list(dictionary.values()))
        
        if fmt == 'pandas':            
            return pd.DataFrame(dictionary, index=[0])
        
        if fmt == 'html':
	    
	    #Create and initialize variable to store station info in html table:            
            html_table = """<meta content="text/html; charset=UTF-8">
                            <style>td{padding: 3px;}</style><table>"""

            #Loop through all keys of station dictionary:
            for k in dictionary.keys():

                #Check if current dict key holds the the long name of the station and
                #if the key to the URL of the station landing page is included:
                if((k=='name') & ('uri' in dictionary.keys())):
                    
                    #Create table row and add link with URL to station landing page:
                    html_table = html_table+'<tr><td>'+k+'</td><td><b><a href="'+str(dictionary['uri'][0])+'"target="_blank">'+str(dictionary[k])+'</a></b></td></tr>'           
                
                #Skip creating table row for 'uri' key:
                elif(k=='uri'):
                    continue
                
                #Dict key 'project' returns a list. Print all comma-sep items as string.
                elif(k=='project'):

                    project_str = ', '.join(dictionary[k])
                    html_table = html_table+'<tr><td>'+k+'</td><td><b>'+project_str+'</b></td></tr>'
                
                #Add table row with station info for current key:
                else:
                    html_table = html_table+'<tr><td>'+k+'</td><td><b>'+str(dictionary[k])+'</b></td></tr>'
            
            #Add html closing tag for table:
            html_table = html_table +'</table>'
            
            #Return station info as html table:
            return html_table
        
    def _setData(self):
        
        """
        Query the sparql endpoint for data products submitted by this station
        adjust latitude and longitude and store a list of data specifications
        and data objects (PID's)
        """
        
        if not self.stationId:
            return
        """      
        # get the ressource url and adjust lat and lon from data portal
        query = sparqls.stationResource(self.stationId)
        key, val = RunSparql(query, 'array').run()
        if val:            
            self.url = val[0][0]
            self.lat = float(val[0][2])
            self.lon = float(val[0][3])
        """            
        
        # it is possible, that a station id has multiple URI
        # ask for all URI
        query = sparqls.stationData(self.uri,'all')
        data = RunSparql(query, 'pandas').run()
        
        if not data.empty:
            self._data = data            
        else:
             self._data = 'no data available'            
         
         # check if data is available and extract the 'unique' data products        
        if isinstance(self._data, pd.DataFrame):
             p = self._data['specLabel'].unique()            
             self._products = pd.DataFrame(p)
             
             # replace samplingheight=None with empty string
             self._data.samplingheight.replace(to_replace=[None], value="", inplace=True)
        else:
             self._products = 'no data available'

            
    def sh(self, product = None):
        """        
        This function is a short cut to return the getSamplingHeight()        
        for documentation, please refer to .getSamplingHeight()
        """
        return self.getSamplingHeight(product)
    
    def getSamplingHeight(self, product = None):
        """        
        a list of unique values for sampling heights for the specified data product
        in case of no sampling hights or the product is not found for this
        station, an empty list is returned.
        A short-cut function is defined .sh() which calls this function.
        
        Parameters
        ----------
        product : str,  digital object specification        

        Returns
        -------
        list, empty if sampling height for product is not found.
        
        """
        # default return empty list
        sh = ['']
        
        # check if product is availabe for station
        if not product in self._data.values:
            return sh
        
        # if product is available but no sampling height is defined, return
        # count returns zero, if no sampling heights found
        if not self._data['samplingheight'][self._data.specLabel.str.match(product)].count():
            return sh
        
        # finally get all sampling heights and create a unique list        
        sh = self._data.samplingheight[self._data.specLabel == product].unique()        
        sh = list(filter(None, sh))
        if not sh:
            return['']
        
        # at this point we have to assume we have an unsorted list of
        # samplingheights as strings. cast to float and sort.
        sh = [float(s) for s in sh]
        sh.sort()
        return sh
        
        
# --EOF Station Class-----------------------------------------            
# ------------------------------------------------------------
def get(stationId):
    """
    Parameters
    ----------
    stationId : str StationId as you can extract with getIdList. Example 'NOR'
                for Atmosphere Norunda station
    example :  get('NOR')
            
    Returns
    -------
    station : object Returns a station object (if stationId is found)

    """
    
    # create the station instance
    myStn = Station()  
    
    query = sparqls.getStations(stationId)    
    stn = RunSparql(query,'pandas').run()
    
    if stn.empty:
        myStn.stationId = stationId
        myStn.valid = False        
        return myStn
    else:
        stn['project'] = stn.apply(lambda x : __project(x['uri']), axis=1)

    # we have found a valid id
    myStn.stationId = stn.id.values[0]
    myStn.valid = True
    
    # it is possible that more than one station has the same id
    # we will give precedence to the icos project
    # but provide a list of URI and projects affiliations
    
    # get lat, lon, eas from icos entry
    if not stn[stn.project.str.upper() == "ICOS"].empty:
        if stn.lat.any():
            myStn.lat = float(stn.lat[stn.project.str.upper() == 'ICOS'])
        if stn.lat.any():
            myStn.lon = float(stn.lon[stn.project.str.upper() == 'ICOS'])
        if stn.elevation.any():
            myStn.eas = float(stn.elevation[stn.project.str.upper() == 'ICOS'])
    else: 
        if stn.lat.any():
            myStn.lat = float(stn.lat.values[0])
        if stn.lon.any():
            myStn.lon = float(stn.lon.values[0])
        if stn.elevation.any():
            myStn.eas = float(stn.elevation.values[0])
        
    
    myStn.name = stn.name.iloc[0]
    myStn.country = stn.country.iloc[0]
    myStn.project = stn.project.tolist()
    myStn.uri = stn.uri.tolist()
        
    
    # if the station belongs to ICOS
    # add information from the labeling app.
    # this is an interim step and should be removed after the full meta data
    # flow from the thematic centres is achieved.        
    
    if 'ICOS' in myStn.project:            
        query = sparqls.stations_with_pi(stationId)
        lstn = RunSparql(query,'pandas').run()
        if not lstn.empty:
            myStn.theme = lstn['stationTheme'].values[0]
            myStn.icosclass = lstn['class'].values[0]
            myStn.icosclass = lstn['class'].values[0]
            myStn.firstName = lstn['firstName'].values[0]
            myStn.lastName = lstn['lastName'].values[0]
            myStn.email = lstn['email'].values[0]
            myStn.siteType = lstn['siteType'].values[0]
            myStn.eag = lstn['eag'].values[0]

    myStn._setData()        
    return myStn

    
def getIdList(project='ICOS', sort='name'):
    """
    Returns a list with all station id's. By default only only ICOS stations
    will be returned. If filter is set to 'all',  all known station id's are retured.
    Please be aware, that the usage of data associated with non-ICOS station
    might be different then the CCBY 4.0 Licence at ICOS.

    Parameters
    ----------
    project : str, optional. The default is 'ICOS'. If you provide the
                set project to "all" for all known stations
                or provide a project name (see getIdList.project)                
                             
    sort : str, optional. The default is 'name'.
    
    Returns
    -------
    Pandas data frame with station id's, name and country.
    """    
    
    project = project.upper()
    
    query = sparqls.getStations()    
    stn = RunSparql(query,'pandas').run()
    stn['project'] = stn.apply(lambda x : __project(x['uri']), axis=1)
    stn['theme'] = stn.apply(lambda x: x['uri'].split('/')[-1].split('_')[0], axis=1)
    
    if not project == 'ALL':
        stn = stn[stn.project == project.upper()]
    
    stn.sort_values(by='name', inplace=True)
    
    return stn
    
def __project(uri):
    """
    internal use, do not call directly. Lambda function
    to apply on dataframe to get a column for 'project'

    Parameters
    ----------
    uri : Carbon Portal resource descriptor
    
    Returns 
    -------
    project : str, from predefined dict

    """
    uri = uri.lower().split('/')[-1].split('_')[0]
    project = {
        'as' : "ICOS",
        'es' : "ICOS",
        'os' : "ICOS", 
        'neon' : 'NEON', 
        'ingos' : 'INGOS' ,   
        'fluxnet' : 'FLUXNET'
        }
    
    if uri in project:
        return project.get(uri)
    else:
        return 'other'
    
def getList(theme=['AS','ES','OS'], ids=None):
    """
    Query the SPARQL endpoint for stations, create an object for each Station
    and return the list of stations.
    By default all (Ocean, Ecosystem, Atmosphere) ICOS stations are returned
    which have been certified (Class 1 & 2). 
    NOTE: if you provide a station list, all other parameters are ignored.
    
    Example:
    # get all ICOS stations 
    getStationList()
    
    # get a list of station objects for ICOS Atmospheric stations
    getStationList('As')

    # get a list for Atmosphere and Ocean stations
    getStationList(['As','OS']) 
    
    # get list of stations with Id's (id's are case sensitive..)
    getStationList(ids=['HTM', 'LMP', SAC])
    
    Parameters
    ----------
    theme : str | [iterable object of strings]) valid entries AS, ES, OS
    
    stationIds : str | [iterable object of strings]) 
    
    Returns
    -------
    stationList : list of station objects
    
    """
    
    stationList = []
    
    # if a list of id's is provided, ignore theme and class.    
    if ids:
        for s in tqdm(ids):
            stationList.append(get(s))
            
        return stationList
    
    defaulttheme=['AS','ES','OS']    
     # make sure input is in uppercase    
    try:        
        if isinstance(theme, str):            
            if not theme.upper() in defaulttheme:
                raise Exception('theme')
        else:
           theme = [x.upper() for x in theme]                  
           
    except:
        # looks like input is not a string and/or not an iterable str.
        # comparison failed. Revert to default values, return all certified stations.
        theme=['AS','ES','OS']
    
    # get station list, by default returns all icos stations
    stations = getIdList()
    # filter by theme
    stations = stations[stations.theme.isin(theme)]
    
    stationList = []
    for s in tqdm(stations.id):
        stationList.append(get(s))
            
    return stationList
# ------------------------------------------------------------    
if __name__ == "__main__":
    """
    execute only if run as a script
    get a full list from the CarbonPortal SPARQL endpoint
    and create a list of station objects    
    """
    msg = """
    You have chosen to run this as a standalone script.
    usually you would try to import a station in your own script
    and then you have the following basic commands [examples]:
    
    # return a list of ICOS Station ID's
    idList = station.getIdList 
    
    # return the station object for ID (example > Norunda in Sweden)
    myStation = station.get('NOR') 
    
    # return a list of station objects for all Ocean ICOS stations.
    stationList = station.getList(['OS']) 
    """
    print(msg)
    

# ------------------------------------------------------------
